use axum::{
    extract::State,
    response::Html,
    routing::{get, post},
    Form, Router,
};
use r2d2_sqlite::rusqlite::{self, Row};
use serde::{Deserialize, Serialize};
use tera::Context;

#[derive(Serialize)]
struct Plant {
    id: i32,
    name: String,
    species: String,
    description: String,
}

impl TryFrom<&Row<'_>> for Plant {
    type Error = rusqlite::Error;

    fn try_from(row: &Row<'_>) -> Result<Self, Self::Error> {
        Ok(Self {
            id: row.get(0)?,
            name: row.get(1)?,
            species: row.get(2)?,
            description: row.get(3)?,
        })
    }
}

pub fn app(ctx: crate::Context) -> axum::Router {
    let mut router = Router::new()
        .route("/", get(get_index))
        .route("/plants", post(post_plants));

    for (name, content) in crate::STATIC_FILES {
        router = router.route(&format!("/static/{}", name), get(content))
    }

    router.with_state(ctx)
}

async fn get_index(State(ctx): State<crate::Context>) -> Html<String> {
    let plants = tokio::task::spawn_blocking(move || {
        let conn = ctx.pool.get().unwrap();

        let mut stmt = conn.prepare("select * from plants").unwrap();
        let mut plants = Vec::new();

        for plant in stmt.query_map((), |row| Plant::try_from(row)).unwrap() {
            plants.push(plant.unwrap());
        }

        plants
    })
    .await
    .unwrap();

    let mut context = Context::new();
    context.insert("plants", &plants);
    Html(ctx.tera.render("index.html", &context).unwrap())
}

#[derive(Deserialize)]
struct NewPlant {
    name: String,
    species: String,
    description: String,
}

async fn post_plants(
    State(ctx): State<crate::Context>,
    Form(plant): Form<NewPlant>,
) -> Html<String> {
    let plant = tokio::task::spawn_blocking(move || {
        let conn = ctx.pool.get().unwrap();

        let mut stmt = conn
            .prepare(
                "insert into plants (name, species, description) values ($1, $2, $3) returning *",
            )
            .unwrap();
        let plant = stmt
            .query_row((&plant.name, &plant.species, &plant.description), |row| {
                Plant::try_from(row)
            })
            .unwrap();

        plant
    })
    .await
    .unwrap();

    let mut context = Context::new();
    context.insert("plant", &plant);
    Html(ctx.tera.render("plant_li.html", &context).unwrap())
}
